/*******************************************************************************
 *******************************************************************************
 **                                                                           **
 **                                                                           **
 **  Copyright 2003-2020 Tom Bascom, Greenfield Technologies                  **
 **  http://www.greenfieldtech.com                                            **
 **                                                                           **
 **  ProTop is free software; you can redistribute it and/or modify it        **
 **  under the terms of the GNU General Public License (GPL) as published     **
 **  by the Free Software Foundation; either version 2 of the License, or     **
 **  at your option) any later version.                                       **
 **                                                                           **
 **  ProTop is distributed in the hope that it will be useful, but WITHOUT    **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    **
 **  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License     **
 **  for more details.                                                        **
 **                                                                           **
 **  See TERMS.TXT for more information regarding the Terms and Conditions    **
 **  of use and alternative licensing options for this software.              **
 **                                                                           **
 **  A copy of the GPL is in GPL.TXT which was provided with this package.    **
 **                                                                           **
 **  See http://www.fsf.org for more information about the GPL.               **
 **                                                                           **
 **                                                                           **
 *******************************************************************************
 *******************************************************************************
 *
 * pro2qmon.p
 *
 *
 * Pro2 replQueue queue monitor
 *
 *
 * Author:
 *
 *	Tom Bascom, Greenfield Technologies
 *	http://www.greenfieldtech.com
 *	August 14, 2019
 *
 */

{lib/protop.i}
{lib/tick.i}
{lib/tt_screenlist.i}

define output parameter dcDescription as character no-undo initial "Pro2Queues".

/*
 *     Pro2 Replication Queue Report              02/17/2021 11:03:41
 *     
 *     Per Thread Queue Summary:
 *     
 *      QNum Oldest Table                     Queue Depth         Age
 *     ----- ------------------------------ -------------    --------
 *         1 n/a                                        0    00:00:00
 *         2 so-trans-d                           468,766    01:26:35
 *         3 n/a                                        0    00:00:00
 *         4 n/a                                        0    00:00:00
 *         5 n/a                                        0    00:00:00
 *     
 *     Per Thread Queue Details:
 *     
 *      QNum Table                            Queue Depth         Age
 *     ----- ------------------------------ -------------    --------
 *         2 so-trans-d                            82,937    01:26:35
 *         2 so-trans-x                            13,360    00:58:23
 *         2 so-trans                             372,469    00:53:27
 *     
 */

define temp-table tt_pro2q no-undo
  field xid             as integer   format ">>9"
  field rq_qNum         as integer   format "->>9"          label "QNum"
  field rq_qName        as character format "x(6)"          label "QName"
  field rq_qStatus      as character format "x(8)"          label "Status"
  field rq_qType        as character format "x(8)"          label "Action"
  field rq_enabledTbls  as integer   format "->>>>>>9"      label "Enabled"
  field rq_disabledTbls as integer   format "->>>>>>9"      label "Disabled"
  field rq_pausedTbls   as integer   format "->>>>>>9"      label "Paused"
  field rq_orphanTbls   as integer   format "->>>>>>9"      label "Orphans"
  field rq_Depth        as int64     format ">,>>>,>>>,>>9" label "Depth"
  field rq_Note         as character format "x(1)"          label ""
  field rq_lagTime      as character format "x(14)"         label "     Queue Lag "
  field rq_oldTblName   as character format "x(30)"         label "Oldest Table"
  field rq_oldTblDB     as character format "x(12)"         label "Source DB"
  field fluff           as character format "x(13)"         label ""
  field zLagTime        as int64
  index xid-idx         is unique primary xid
  index rq_Num-idx      is unique rq_qNum
.

define temp-table tt_replQueue no-undo
  field qNum          as integer   format "->>>9"         label "QNum"
  field qStatus       as character format "x(8)"          label "Status"
  field qType         as character format "x(8)"          label "Type"
  field srcDB         as character format "x(12)"         label "DB"
  field tblName       as character format "x(30)"         label "Table"
  field tblStatus     as character format "x(10)"         label "Status"
  field qDepth        as int64     format ">,>>>,>>>,>>9" label "Queue Depth"
  field oldest        as datetime
  index qNum-idx is primary unique qNum tblName
.


/* return seconds as a string with format:  #days hh:mm:ss
 */

function ageString returns character ( s as int64 ):

  define variable d  as int64 no-undo.
  define variable s1 as int64 no-undo.

  if s = ? then
    return "".
   else
    do:
      d = int64( truncate( s / 86400, 0 )).
      s = s modulo 86400.
      return string( d, ">>>>>" ) + " " + string( s, "hh:mm:ss" ).
    end.

end.


/* someone thought that time should be stored as a formatted string, let's fix that!
 */

function stringTime2seconds returns integer ( st as character ):
  return
    ( integer( entry( 1, st, ":" )) * 3600 ) +
    ( integer( entry( 2, st, ":" )) * 60 ) +
    ( integer( entry( 3, st, ":" )))
  .
end.


define variable hasPro2 as logical no-undo.

define variable b_rq    as handle  no-undo.
define variable b_rc    as handle  no-undo.
define variable b_tbl   as handle  no-undo.

define variable q_rq    as handle  no-undo.
define variable q_tbl   as handle  no-undo.

procedure mon-init:

  define variable b as handle.

  b = ( temp-table tt_pro2q:default-buffer-handle ).

  if pt_altWidth = "80" then
    do:
      altFormat( "80", b, "rq_qNum",         "->>9" ).
      altFormat( "80", b, "rq_qType",        "x(6)" ).
      altFormat( "80", b, "rq_enabledTbls",  ">>>>>9" ).
      altFormat( "80", b, "rq_disabledTbls", ">>>>>9" ).
      altFormat( "80", b, "rq_pausedTbls",   ">>>>>9" ).
      altFormat( "80", b, "rq_orphanTbls",   ">>>>>9" ).
      altFormat( "80", b, "rq_depth",        ">>>>>9" ).
      altFormat( "80", b, "rq_Note",         "x(1)" ).
      altFormat( "80", b, "rq_lagTime",      "x(14)" ).
      altFormat( "80", b, "rq_oldTblName",   "x(17)" ).
    end.

  find dictdb._File no-lock where _File-Name = "replQueue" no-error.
  hasPro2 = available( _File ).

  if hasPro2 = no then
    return.

  create buffer b_rc  for table "dictdb.replControl".
  create buffer b_rq  for table "dictdb.replQueue".
  create buffer b_tbl for table "dictdb.replTableXref".

  create query q_rq.
  q_rq:set-buffers( b_rq ).
  q_rq:query-prepare( substitute( 'preselect each dictdb.replQueue no-lock where applied = no and sequence >= &1', pt_pro2qSkip2Seq )).

  create query q_tbl.
  q_tbl:set-buffers( b_tbl ).
  q_tbl:query-prepare( "preselect each dictdb.replTableXref no-lock" ).	/* really ought to add srcDB?	*/

  return.

end.


procedure mon-update:

  define input parameter argList as character no-undo.

  define variable i  as integer no-undo.
  define variable n  as integer no-undo.
  define variable ok as logical no-undo.

  define variable q1     as int64     no-undo.
  define variable q2     as int64     no-undo.
  define variable qd     as int64     no-undo.
  define variable qn     as int64     no-undo.
  define variable qz     as int64     no-undo.

  define variable inQ    as int64     no-undo format ">,>>>,>>>,>>9" label "Queue Depth".
  define variable oldQ   as datetime  no-undo                        label "Age".
  define variable oldTbl as character no-undo format "x(30)"         label "Oldest Table".

  define variable oldTblName as character no-undo.

  if hasPro2 = no then
    do:

      find tt_screenlist where tt_screenlist.dcName = dcDescription no-error.

      if available( tt_screenlist ) and tt_screenlist.screenVisible = yes then
        do:

          message
            "This database does not contain a Pro2 Replication Schema."
           view-as
            alert-box information
          .

          publish "protop_show" ( "pro2queues" ).		/* disable the screen	*/

        end.

      return.

    end.

  empty temp-table tt_pro2q.
  empty temp-table tt_replQueue.

  /* build the list of tables being replicated (maybe only do this every X iterations? it shouldn't change very often)
   */

  q_tbl:query-open.
  do while q_tbl:get-next( no-lock ):

    create tt_replQueue.
    assign
      tt_replQueue.qNum    = b_tbl:buffer-field( "qThread" ):buffer-value()
      tt_replQueue.srcDB   = b_tbl:buffer-field( "srcDB" ):buffer-value()
      tt_replQueue.tblName = b_tbl:buffer-field( "srcTable" ):buffer-value()
      tt_replQueue.oldest  = datetime( today + 1, 0 )					/* intialize to "tomorrow" for use with min(), below		*/
      n = max( n, tt_replQueue.qNum )							/* Pro2 queue numbers might have holes in them			*/
    .

    if b_tbl:buffer-field( "genQrec" ):buffer-value() = false then
      tt_replQueue.tblStatus = "disabled".
     else
      do:
        tt_replQueue.tblStatus = "enabled".
        if b_tbl:buffer-field( "procQrec" ):buffer-value() = false then			/* if it isn't "enabled" then "paused" doesn't make sense	*/
          tt_replQueue.tblStatus = "paused".
      end.

  end.  
  q_tbl:query-close.

  assign
    q1 = 0
    q2 = dynamic-current-value( "NextReplNbr", "dictdb" ) - 1
    qd = 0
    qn = 0
    qz = 0
  .

  do i = 1 to n:

    /* calculate the queue depth by looking at the first record and the next sequence#
     * this can be fooled if there are orphan records in the replQueue
     */

    ok = b_rq:find-first( substitute( "where applied = no and qThread = &1 and sequence > &2", i, pt_pro2qSkip2Seq ), no-lock ) no-error.
    if ok then
      do:

        qn = int64( b_rq:buffer-field( "sequence" ):buffer-value()).

        if qn <> ? then
          do:
            if q1 = 0 then
              q1 = qn.
             else
              q1 = min( q1, qn ).
          end.

      end.

    if qn <> ? and qn <> 0 then
      do:
        qz = qz + ( q2 - qn ).
      end.

  end.

  if q1 > 0 then
    assign
      qd = q2 - q1
    .

  /* gather metrics by counting every replQueue record - this is very painful if Pro2 falls behind :( so only do it if the queue depth appears small
   */

  if qd < pt_pro2qCountLim then
    do:

      q_rq:query-open.
      do while q_rq:get-next( no-lock ):	/* for each replQueue where applied = no and sequence > pt_pro2qSkip2Seq...	*/

        find tt_replQueue where tt_replQueue.qNum = b_rq:buffer-field( "qThread" ):buffer-value() and tt_replQueue.tblName = b_rq:buffer-field( "srcTable" ):buffer-value() no-error.

        if available tt_replQueue then
          assign
            tt_replQueue.qDepth = tt_replQueue.qDepth + 1
            tt_replQueue.oldest = min( tt_replQueue.oldest, datetime( b_rq:buffer-field( "eventDate" ):buffer-value(), stringTime2seconds( b_rq:buffer-field( "eventTime" ):buffer-value() ) * 1000 ))
          .

      end.
      q_rq:query-close.

    end.

  /* summarize the details for each defined replQueue
   */

  do i = 1 to n:

    find first tt_replQueue where qNum = i no-error.
    if not available tt_replQueue then next.						/* Pro2 skipped a queue#					*/

    create tt_pro2q.
    assign
      tt_pro2q.xid      = i
      tt_pro2q.rq_qNum  = i
      tt_pro2q.rq_qName = substitute( "q&1", trim( string( i, "-999" )))
    .

    ok = b_rc:find-first( substitute( 'where groupId = "control" and codeId = "replication" and codeVal1 = "&1"', i ), no-lock ) no-error.
    tt_pro2q.rq_qStatus = ( if ok then b_rc:buffer-field( "codeVal3" ):buffer-value() else "?" ).

    if tt_pro2q.rq_qStatus = "1" then
      tt_pro2q.rq_qStatus = "Running".	/* no other values are known...	*/

    ok = b_rc:find-first( substitute( 'where groupId = "queue" and codeId = "disposition" and codeVal1 = "replication" and codeVal2 = "&1"', i ), no-lock ) no-error.
    tt_pro2q.rq_qType = ( if ok then b_rc:buffer-field( "codeVal3" ):buffer-value() else "?" ).

    if tt_pro2q.rq_qType = "D" then
      tt_pro2q.rq_qType = "Delete".
     else if tt_pro2q.rq_qType = "M" then
      tt_pro2q.rq_qType = "Mark".
     else if tt_pro2q.rq_qType = "A" then
      tt_pro2q.rq_qType = "Audit".		/* theoretical, so far it has never been seen in action...	*/

    inQ = 0.

    for each tt_replQueue where qNum = i:

      if       tt_replQueue.tblStatus = "enabled"  then tt_pro2q.rq_enabledTbls  = tt_pro2q.rq_enabledTbls  + 1.
       else if tt_replQueue.tblStatus = "disabled" then tt_pro2q.rq_disabledTbls = tt_pro2q.rq_disabledTbls + 1.
       else if tt_replQueue.tblStatus = "paused"   then tt_pro2q.rq_pausedTbls   = tt_pro2q.rq_pausedTbls   + 1.
       else                                             tt_pro2q.rq_orphanTbls   = tt_pro2q.rq_orphanTbls   + 1.

      if tt_replQueue.qDepth <> ? then
        inQ = inQ + tt_replQueue.qDepth.

    end.

    tt_pro2q.rq_note = "x".

    if qd >= pt_pro2qCountLim then
      do:

        tt_pro2q.rq_note = "e".

        /* get queue depth from replControl (only updates occasionally)
         */

        inQ = 0.
        ok = b_rc:find-first( substitute( 'where groupId = "replication" and codeId = "reccnt" and codeVal1 = "&1"', i ), no-lock ) no-error.
        if ok and num-entries( b_rc:buffer-field( "codeVal2" ):buffer-value()) = 2 then
          do:
            tt_pro2q.rq_note = "2".
            inQ = int64( b_rc:buffer-field( "codeVal3" ):buffer-value()).
          end.

        /* estimate queue depth by pro-rating 1st record (q1) vs next seq# (q2)
         */

        if pt_pro2qEstimate = yes then
          do:

            q1 = 0.
            ok = b_rq:find-first( substitute( "where applied = no and qThread = &1 and sequence > &2", i, pt_pro2qSkip2Seq ), no-lock ) no-error.
            if ok then
              assign
                q1 = int64( b_rq:buffer-field( "sequence" ):buffer-value())
                inQ = int64( qd * (( q2 - q1 ) / qz ))
                tt_pro2q.rq_note = "p"
              .

          end.

      end.

    ok = b_rq:find-first( substitute( "where applied = no and qThread = &1 and sequence > &2", i, pt_pro2qSkip2Seq ), no-lock ) no-error.

    if ok = no then
      assign
        tt_pro2q.rq_oldTblName = "n/a"
        tt_pro2q.rq_oldTblDB   = ""
        tt_pro2q.rq_lagTime    = ""
        tt_pro2q.rq_depth      = 0
        tt_pro2q.zlagTime      = 0
        inQ                    = 0
      .
     else
      do:
        assign
          tt_pro2q.zLagTime = interval( now, datetime( b_rq:buffer-field( "eventDate" ):buffer-value(), stringTime2seconds( b_rq:buffer-field( "eventTime" ):buffer-value() ) * 1000 ), "seconds" )
          tt_pro2q.rq_oldTblName = b_rq:buffer-field( "srcTable" ):buffer-value()
          tt_pro2q.rq_oldTblDB   = b_rq:buffer-field( "srcDB" ):buffer-value()
        .
        find tt_replQueue where tt_replQueue.qNum = i and tt_replQueue.tblName = b_rq:buffer-field( "srcTable" ):buffer-value() no-error.
        if not available tt_replQueue then
          tt_pro2q.rq_orphanTbls   = tt_pro2q.rq_orphanTbls + 1.
      end.

    assign
      tt_pro2q.rq_depth      = ( if inQ < 0 then 1 else inQ )	/* negative values are probably bad estimates from proportional calc so 1 is better than 0	*/
      tt_pro2q.rq_lagTime    = ageString( tt_pro2q.zLagTime )
    .

  end.

  add2ds( temp-table tt_pro2q:default-buffer-handle ).

  return.

end.

return.
