/* clife.p
 *
 * Conway's Game of Life 
 *
 * https://en.wikipedia.org/wiki/Conway%27s_Game_of_Life
 *
 *	Any live cell with fewer than two live neighbours dies, as if by underpopulation.
 *	Any live cell with two or three live neighbours lives on to the next generation.
 *	Any live cell with more than three live neighbours dies, as if by overpopulation.
 *	Any dead cell with exactly three live neighbours becomes a live cell, as if by reproduction.
 *
 * https://xkcd.com/2293/ (RIP John Conway)
 *
 * pro -p clife.p -param "seedfile.clife"
 *
 * the seed file is lines of text where lines that begin with "!" or "#" are
 * comments, "x", "o", or "1" indicates a live cell, any other character a
 * dead cell.
 *
 * this file format is very similar to:  https://www.conwaylife.com/wiki/Plaintext
 *
 *	    xxx
 * 	    x x
 *	    x x
 *	     x
 *	  x xxx
 *	   x x x
 *	     x  x
 *	    x x
 *	    x x
 */

&global-define	WIDTH	128

define temp-table grid no-undo
  field r as integer							/* row #							*/
  field a as integer extent {&WIDTH}					/* live row data						*/
  field b as integer extent {&WIDTH}					/* scratch row data						*/
  field s as character format "x({&WIDTH})"				/* string holding the display values				*/
  index r-idx is unique primary r
.

define variable w as integer no-undo initial {&WIDTH}.			/* screen width							*/
define variable h as integer no-undo.					/* screen height						*/

define variable e as logical no-undo.					/* use emojis?	(just for fun really)				*/

define variable gen as integer no-undo.					/* generation							*/
define variable y   as integer no-undo.					/* current population						*/
define variable z   as integer no-undo.					/* max population						*/

define variable r1 as integer no-undo extent.
define variable r2 as integer no-undo extent.
define variable r3 as integer no-undo extent.

define variable i as integer no-undo.
define variable j as integer no-undo.

/* random mutation support ;)
 */

define variable lo  as integer no-undo initial 2.			/* lo random limit, must be <= 1 for randomness to happen	*/
define variable hi  as integer no-undo initial 5000.			/* hi random limit, must be > lo				*/

define variable ex  as character no-undo initial ".".

define variable fileName      as character no-undo.
define variable return_status as logical   no-undo.

define stream inStrm.

h = screen-lines - 6.

form
  /* grid.r format "999" */ grid.s skip
 with
  frame gs
  title " Conway's Game of Life "
  no-labels
  overlay
  h down
  row 4
  column 16
  width 132
.

run seedGrid( session:parameter ).

/* iterate generations - q to quit, p to pause
 */

do while true:

  run showGrid.

  z = max( y, z ).								/* track max population						*/

  put screen row screen-lines + 2 column 1 substitute( " &1 &2 &3   ", gen, y, z ).
  put screen row screen-lines + 2 column 70 'press "p" to pause, "q" to quit, "r" to randomly reseed, "f" to choose file.'.

  run nextGen.									/* calculate the next generation				*/
  gen = gen + 1.

  readkey pause 0.1.								/* increase pause value to slow things down     		*/

  if lastkey = asc( "q" ) then							/* quit!							*/
    leave.
   else if lastkey = asc( "p" ) then						/* pause updates until a key is pressed				*/
    readkey.
   else if lastkey = asc( "c" ) then						/* change the drawing character					*/
    do:
      readkey.
      if lastkey > 32 and lastkey < 127 then
        ex = chr( lastkey ).
    end.
   else if lastkey = asc( "e" ) then						/* toggle "use emojis"						*/
    e = not( e ).
   else if lastkey = asc( "r" ) then						/* reseed with something random					*/
    do:
      do i = 1 to h:
        find grid where grid.r = i.
        do j = 1 to w:
          if random( 0, 9 ) = 0 then						/* 10% chance of being alive					*/
            grid.a[j] = 1.
           else
            grid.a[j] = 0.
        end.
      end.
      next.
    end.
   else if lastkey = asc( "f" ) then						/* reseed from a file						*/
    do:
      run lib/filepicker.p ( "etc", "*.clife", input-output fileName ).
      file-info:file-name = fileName.
      if file-info:full-pathname <> ? then run seedGrid( fileName ).
      next.
    end.

  if y = 0 and lo > 1 then							/* if there is no randomness then mutations			*/
    do:										/* cannot occur thus y = 0 means every cell			*/
      pause.									/* is dead and we are done					*/	
      leave.
    end.

end.

hide frame gs.

readkey pause 0.

return.

/* end of main block
 */


/* initialize the grid by reading a seed file, if there is no seed file create something fun
 */

procedure seedGrid:

  define input parameter seedFileName as character no-undo.

  define variable i as integer no-undo.
  define variable j as integer no-undo.
  define variable n as integer no-undo.

  define variable c     as character no-undo.
  define variable xLine as character no-undo.

  empty temp-table grid.

  do i = 1 to h:							/* create a blank grid   */
    create grid.
    assign
      grid.r = i
      grid.a = 0
      grid.b = 0
    .
  end.

  file-info:file-name = seedFileName.
  if file-info:full-pathname = ? then
    do:

      find grid where grid.r = ( h - 10 ).			  /* if there is no seed file use https://xkcd.com/2293/ (RIP John Conway)	*/
      grid.a[5] = 1.  grid.a[6] = 1.  grid.a[7] = 1.
      find next grid.
      grid.a[5] = 1.  grid.a[7] = 1.
      find next grid.
      grid.a[5] = 1.  grid.a[7] = 1.
      find next grid.
      grid.a[6] = 1.
      find next grid.
      grid.a[3] = 1.  grid.a[5] = 1.  grid.a[6] = 1.  grid.a[7] = 1.
      find next grid.
      grid.a[4] = 1.  grid.a[6] = 1.  grid.a[8] = 1.
      find next grid.
      grid.a[6] = 1.  grid.a[9] = 1.
      find next grid.
      grid.a[5] = 1.  grid.a[7] = 1.
      find next grid.
      grid.a[5] = 1.  grid.a[7] = 1.

      run showGrid.
      pause.

    end.
   else										/* load a seed file into the grid				*/
    do:

      input stream inStrm from value( file-info:full-pathname ).

      i = 0.
      repeat:

        xLine = "".
        import stream inStrm unformatted xLine.

        if xLine begins "!" or xLine begins "#" then				/* comment lines						*/
          next.

        i = i + 1.
        find grid where grid.r = i no-error.
        if i > h or not available grid then leave.				/* ignore data beyond the defined height			*/

        n = minimum( w, length( xLine )).					/* don't process more than the screen width			*/
        do j = 1 to n:
          c = substring( xLine, j, 1 ).
          if index( "xo1", c ) > 0 then						/* x, o, or 1 are alive - not case sensitive			*/
            grid.a[j] = 1.
        end.

      end.
      input stream inStrm close.

    end.

  return.

end.


/* check the neighborhood population
 */

function cellStatus returns integer ( r1 as integer extent, r2 as integer extent, r3 as integer extent, c as integer ):

  define variable cl as integer no-undo.					/* neighbor column left						*/
  define variable cr as integer no-undo.					/* neighbor column right					*/

  define variable x  as integer no-undo.					/* current cell							*/
  define variable p  as integer no-undo.					/* population of neighbors					*/

  /* precompute wrapped column numbers
   */

  cl = c - 1.
  if cl < 1 then cl = w.							/* wrap to the end						*/
  cr = c + 1.
  if cr > w then cr = 1.							/* wrap to the beginning					*/

  /* count the neighboring population - but don't count yourself! (r2[c])
   *
   */

  p = r1[cl] +      r1[c]       + r1[cr] +
      r2[cl] +  /** r2[c] **/     r2[cr] +
      r3[cl] +      r3[c]       + r3[cr]
  .

  x = r2[c].									/* state of the current cell					*/

  /* x = 1 = live, x = 0 = dead
   *
   *	Any live cell with fewer than two live neighbours dies, as if by underpopulation.
   *	Any live cell with two or three live neighbours lives on to the next generation.
   *	Any live cell with more than three live neighbours dies, as if by overpopulation.
   *
   *	Any dead cell with exactly three live neighbours becomes a live cell, as if by reproduction.
   */

  if (( x = 1 ) and ( p = 2 or p = 3 )) or
     (( x = 0 ) and ( p = 3 )) then
    do:
      if random( lo, hi ) = 1 then
        return 0.								/* random mutation!						*/
       else
        return 1.								/* normal behavior						*/
    end.
   else 
    do:
      if random( lo, hi ) = 1 then
        return 1.								/* random mutation!						*/
       else
        return 0.								/* normal behavior						*/
    end.

end.


/* calculate the next generation (grid.b) based on the
 * population density of the current generation (grid.a)
 */

procedure nextGen:

  define variable i  as integer no-undo.				/* row								*/
  define variable j  as integer no-undo.				/* column							*/

  define variable li as integer no-undo.				/* left index							*/
  define variable ri as integer no-undo.				/* right index							*/
 
  define variable z  as character no-undo.				/* blank row detection						*/

  define variable nr as integer no-undo.				/* neighbor row							*/

  /* compute the next generation into the scratch rows using
   * a sliding window of three rows (i - 1, i, i + 1)
   */

  do i = 1 to h:

    assign
      z = ""								/* assume a blank row						*/
      li = 99999							/* left index starts out waaaaay to the right			*/
      ri = -1								/* right index starts to the left				*/
    .

    nr = i - 1.
    if nr < 1 then nr = h.						/* wrap around to the last row					*/
    find grid where grid.r = nr no-error.

    r1 = grid.a.
    z = grid.s.								/* add the row							*/
    li = minimum( li, index( grid.s, ex )).				/* the leftmost live cell					*/
    ri = maximum( ri, r-index( grid.s, ex )).				/* the rightmost live cell					*/

    nr = i + 1.
    if nr > h then nr = 1.						/* wrap around to the 1st row					*/
    find grid where grid.r = nr no-error.

    r3 = grid.a.
    z = z + grid.s.							/* add the row							*/
    li = minimum( li, index( grid.s, ex )).				/* the leftmost live cell					*/
    ri = maximum( ri, r-index( grid.s, ex )).				/* the rightmost live cell					*/

    find grid where grid.r = i.						/* find the current grid line last so				*/
									/* that we can update the current cell				*/
    r2 = grid.a.
    z = z + grid.s.							/* add the row							*/
    li = minimum( li, index( grid.s, ex )).				/* the leftmost live cell					*/
    ri = maximum( ri, r-index( grid.s, ex )).				/* the rightmost live cell					*/

    li = maximum( 1, li - 1 ).						/* look one cell to the left					*/
    ri = minimum( w, ri + 1 ).						/* look one cell to the right					*/

    if ri = w then li = 1.						/* edge wrapping is not optimized				*/
    if li = 1 then ri = w.						/* edge wrapping is not optimized				*/

    /* stupid 4gl trick warning!
     *
     * a string with all spaces = blank as far as 4gl string
     * comparisons care so we can easily test if all 3 lines
     * are blank and then skip it!
     */

    if z <> "" then							/* skip multiple empty lines					*/
      do j = li to ri:							/* skip blocks of empty columns					*/
        grid.b[j] = cellStatus( r1, r2, r3, j ).			/* populate the new layer					*/
      end.

  end.

  /* copy the scratch rows to the live rows
   */

  for each grid:
    grid.a = grid.b.
  end.

end.


{lib/emoji.i}

procedure showGrid:

  define variable j as integer no-undo.

  y = 0.								/* current live population					*/

  for each grid:
    grid.s = "".							/* string display of live vs dead				*/
    do j = 1 to w:
      assign
        y = y + grid.a[j]
        grid.s = grid.s + ( if grid.a[j] = 0 then " " else ( if e = no then ex else emoji[ random( 1, 16 ) ] )).
      .
    end.
    display /* grid.r */ grid.s with frame gs.
    down with frame gs.
  end.

end.
